//
//  KFDecorator.h v1.0
//  KFStandard
//
//  Created by Ken Ferry on 1/20/05.
//  Copyright 2005 Ken Ferry. All rights reserved.
//
//  Introduction and usage:
//  
//  KFDecorator is a class that uses FScript blocks to produce general purpose 
//  decorator objects. I don't use it in release code, but I've found it to be
//  very useful when using FScript (Anywhere) as part of the development
//  environment for a cocoa application.  It's good for debugging,
//  exploration of ideas, and automated or interactive testing.  
//
//  The idea of a decorator is that it wraps a base object and modifies
//  its behavior.  The decorator catches specified messages and processes
//  them, and any messages it doesn't catch are forwarded to the base
//  object.  It's a design pattern from the Gang of Four book.
//  
//  Examples are a good way to understand this class.  The following FScript
//  snippet would help you to understand the behavior of NSTableView by beeping 
//  every time a table sends -numberOfRowsInTableView:.  Suppose that tableView is
//  an instance of NSTableView and dataSource is its data source.
// 
//        > dataDecorator := KFDecorator decoratorWithBase:dataSource
//
//        > dataDecorator setBlock:[:self :tv | sys beep. self base numberOfRowsInTableView:tv] forSelector:#numberOfRowsInTableView:
//
//        > dataDecorator
//        <KFDecorator[0x871f640] base:<TableViewManager: 0x4e68ae0> maps:{
//             #numberOfRowsInTableView: => [:self :tv | sys beep. self base numberOfRowsInTableView:tv]
//        }>
//
//        > tableView setDataSource:dataDecorator
//
//        > dataSource numberOfRowsInTableView:tableView
//        1102
//
//        > dataDecorator numberOfRowsInTableView:tableView
//        1102
//  
//  Except for the beeping, the table will behave as it did before.  Note the use of 'self base' in the last line.  Calls to
//  the base object in a decorator are similar to calls to super in a normal subclassing situation. 
//
//  You can use KFDecorator instances that have no base.  In this case, a decorator object is an object whose interface and
//  implementation is completely determined at runtime.  The object is a bundled up collection of FScript blocks that can be called 
//  with objective-C messages.
//
//  This can be useful in testing (automated or interactive) situations.  You can use a decorator to inline define 'mock' 
//  objects of whatever level of complexity (or simplicity) is required.  Here's how you might mock the datasource for a table:
//
//        > tableDataSource := KFDecorator decorator
//
//        > tableDataSource setBlock:[:self :tv | 10] 
//            forSelector:#numberOfRowsInTableView:
//            withSignatureTypes:'i@:@'
//
//        > tableDataSource setBlock:[:self :tv :col :row | 'test']
//            forSelector:#tableView:objectValueForTableColumn:row:
//            withSignatureTypes:'@@:@@i'
//
//        > tableDataSource
//        <KFDecorator[0x597a990] base:(null) maps:{
//             #numberOfRowsInTableView: => [:self :tv | 10]
//             #tableView:objectValueForTableColumn:row: => [:self :tv :col :row | 'test']
//        }>
//
//  This code is in FScript, but you can 
//  do the same thing from objective-c using -[NSString(FSNSString) asBlock] to construct the block arguments. 
// 
//  Note that we specified method signatures.  If you do not specify
//  a method signature for a block-selector pair, then arguments and returns are typed:
//       (1) the same way the underlying base object types them, if the base object also responds to the selector
//       (2) as objects, otherwise
//
//  You can also use KFDecorator to switch between existing objective-c
//  methods that you'd like to evaluate.
//
//     decorator setBlock:#doStuff_alternateAlgorithm forSelector:#doStuff
//
//  Limitations:
//
//  You cannot set a block for a method implemented by KFDecorator itself.  This is because KFDecorator
//  uses Objective-C's message forwarding to do its dirty work, and that system is only invoked for 
//  messages that an object does not directly implement.  Please let me know if this is a serious issue for you.
//
//  KFDecorator inherits from NSObject, not NSProxy.  In conjunction with the above, this limits the selectors
//  which may be assigned blocks.  NSProxy lacks a working basic methodSignatureForSelector:, and the only
//  work arounds I know of are pretty nasty.
//
//  KFDecorator uses a private method of FScript's Pointer.h, and a private method of NSMethodSignature 
//  (though I believe the latter is part of the OpenStep spec.  go figure.)  The basic idea is this: KFDecorator is 
//  primarily intended to be part of your development cycle, not part of your release app.
//


#import <Cocoa/Cocoa.h>

@class Block;

@interface KFDecorator : NSObject 
{
    id base;
    NSMutableDictionary *selectorsToBlocks;
    NSMutableDictionary *selectorsToSignatures;
}

-(id)initWithBase:(id)anObject;
+(id)decorator;
+(id)decoratorWithBase:(id)anObject;

-(id)base;
-(void)setBase:(id)anObject;

-(Block *)blockForSelector:(SEL)aSelector;
-(void)setBlock:(Block *)block forSelector:(SEL)aSelector;
-(void)setBlock:(Block *)block forSelector:(SEL)aSelector withSignatureTypes:(NSString *)objCTypes; // e.g. @"I@:@" means "returns unsigned, takes object (implicit self), selector (implicit _cmd), object (type of the 'first' argument from the usual point of view) 
-(void)setBlock:(Block *)block forSelector:(SEL)aSelector withSignature:(NSMethodSignature *)selectorSignature;

@end


@interface NSObject (KFDecorator)

- (id)decorator;

@end
