//
//  KFDecorator.m v1.0
//  KFStandard
//
//  Created by Ken Ferry on 1/20/05.
//  Copyright 2005 Ken Ferry. All rights reserved.
//

#import "KFDecorator.h"
#import <FScript/Pointer.h>
#import <FScript/Block.h>
#import <FScript/FSNSString.h>
#import <FScript/FScriptFunctions.h>

@implementation NSObject (KFDecorator)

- (id)decorator
{
    return [KFDecorator decoratorWithBase:self];
}

@end

@interface Pointer (FromPointerPrivate)

+(Pointer *)pointerWithCPointer:(void *)p type:(const char *)t;

@end

@interface NSMethodSignature (FromNSMethodSignaturePrivate)

+(NSMethodSignature *)signatureWithObjCTypes:(const char *)objCTypes;

@end


@implementation KFDecorator

#pragma mark -
#pragma mark INIT/DEALLOC
#pragma mark -

// designated initializer
-(id)initWithBase:(id)anObject
{
    self = [super init];
    if (self != nil)
    {
        selectorsToBlocks     = [[NSMutableDictionary alloc] init];
        selectorsToSignatures = [[NSMutableDictionary alloc] init];
        [self setBase:anObject];
    }
    
    return self;
}

-(void)dealloc
{
    [self setBase:nil];
    [selectorsToSignatures release];
    [selectorsToBlocks release];
    [super dealloc];
}

#pragma mark 
-(id)init
{
    return [self initWithBase:nil];
}

+(id)decorator
{
    return [[[[self class] alloc] init] autorelease];
}

+(id)decoratorWithBase:(id)anObject
{
    return [[[[self class] alloc] initWithBase:anObject] autorelease];
}


#pragma mark -
#pragma mark PRIVATE/OVERRIDES
#pragma mark -

-(NSString *)description
{
    NSArray *selectorStrings = [[@"[:selVals | NSString stringWithUTF8String:@(selVals pointerValue)]" asBlock] value:[selectorsToBlocks allKeys]];
    NSString *mapLines = [[@"[:selStrings :blocks | ('\n     #' ++ @ selStrings @ ++ ' => '@ ++ @(blocks @description)) \\#++ ++ '\n']" asBlock] value:selectorStrings 
                                                                                                                                                 value:[selectorsToBlocks allValues]];
    return [NSMutableString stringWithFormat:@"<%@[%p] base:%@ maps:{%@}>", [self class], self, base, mapLines];
}

-(int)argumentCountForSelector:(SEL)selector
{
    int count = 0;
    NSString *selectorString = NSStringFromSelector(selector);
    int selectorStringLength = [selectorString length];
    
    NSRange searchRange = NSMakeRange(0,selectorStringLength);
    NSRange resultRange;
    while ((resultRange = [selectorString rangeOfString:@":" options:0 range:searchRange]).location != NSNotFound)
    {
        count++;
        searchRange.location = NSMaxRange(resultRange);
        searchRange.length   = selectorStringLength - searchRange.location;
    }
    
    return count;
}

-(SEL)valueSelectorWithArgCount:(int)argumentCount
{
    if (argumentCount == 0)
    {
        return @selector(value);
    }
    
    NSMutableString *outSelString = [NSMutableString string];    
    
    int i;
    for (i = 0; i < argumentCount; i++)
    {
        [outSelString appendString:@"value:"];
    }
    
    return NSSelectorFromString(outSelString);
}



#pragma mark 

- (BOOL)respondsToSelector:(SEL)aSelector
{
    return ([super respondsToSelector:aSelector] 
            || ([self blockForSelector:aSelector] != nil)
            || [base respondsToSelector:aSelector]);
}


// signature selection priority:
//   signature of a KFDecorator method
//   explicitly specified signature in selectorsToSignatures
//   signature of a base object method
//   everything-is-an-object signature if a block handles aSelector
- (NSMethodSignature *)methodSignatureForSelector:(SEL)aSelector
{
    NSMethodSignature *signature = [super methodSignatureForSelector:aSelector];
    
    if (signature == nil)
    {
        signature = [selectorsToSignatures objectForKey:[NSValue value:&aSelector withObjCType:@encode(SEL)]];
    }
    if (signature == nil)
    {
        signature = [base methodSignatureForSelector:aSelector];
    }
    if (signature == nil)
    {
        // a bit of fishiness here - the argument count for valueSelector may not be the same as 
        // the block's argument count.  The block returns the appropriate all-objects signature anyway.
        SEL valueSelector = [self valueSelectorWithArgCount:[self argumentCountForSelector:aSelector]];
        signature = [[self blockForSelector:aSelector] methodSignatureForSelector:valueSelector];
    }    
    
    return signature;
}


- (void)forwardInvocation:(NSInvocation *)invocation
{
    SEL selector = [invocation selector];
    NSMethodSignature *signature = [invocation methodSignature];
    
    Block *block = [self blockForSelector:selector];
    if (block != nil) // decorated method
    {
        // setup
        /* buffer is a scratch pad big enough to hold any argument or return */
        void *buffer = malloc(MAX([signature frameLength], [signature methodReturnLength])); 
        NSNumber *zeroNum = [NSNumber numberWithInt:0];
        
        NSMutableArray *blockArguments = [Array array]; // Array class allows adding nil to array
        
        // first argument is 'self'
        [blockArguments addObject:self];
                
        // need to convert arguments, which may be c types, to the objects an FScript 
        // Block expects. The Pointer class handles the conversions
        int argIndex, numArguments;
        numArguments = [signature numberOfArguments];

        for (argIndex = 2; argIndex < numArguments; argIndex++)
        {
            const char *argType = [signature getArgumentTypeAtIndex:argIndex];
            [invocation getArgument:buffer atIndex:argIndex];
            Pointer *fs_buffer = [Pointer pointerWithCPointer:buffer type:argType]; // private Pointer method
            id argumentAsObject = [fs_buffer at:zeroNum];
            [blockArguments addObject:argumentAsObject];
        }
        
        id resultAsObject = [block valueWithArguments:blockArguments];
        
        // convert the result to whatever type it's supposed to be
        const char *returnType = [signature methodReturnType];
        /* encoded types are not always uniqued.  Any two @encode(type) are the same in a given bundle, but not in a whole process */
        if (*returnType != *@encode(void))
        {
            // Pointer converts the auxReturnValue object to the C type that the original
            // caller expects
            Pointer *fs_buffer = [Pointer pointerWithCPointer:buffer type:returnType];
            [fs_buffer at:zeroNum put:resultAsObject];
            [invocation setReturnValue:buffer];
        }
        
        // cleanup
        free(buffer);
    }
    else if ([base respondsToSelector:selector]) // undecorated method
    {
        [invocation invokeWithTarget:base];
    }
    else // unhandled method
    {
        [super forwardInvocation:invocation];
    }
}


#pragma mark -
#pragma mark ACCESSORS
#pragma mark -

-(id)base
{
    return base;
}

-(void)setBase:(id)anObject
{
    [anObject retain];
    [base release];
    base = anObject;
}

-(Block *)blockForSelector:(SEL)aSelector
{
    return [selectorsToBlocks objectForKey:[NSValue value:&aSelector withObjCType:@encode(SEL)]];
}

-(void)setBlock:(Block *)block forSelector:(SEL)aSelector
{
    [self setBlock:block forSelector:aSelector withSignature:nil];
}

-(void)setBlock:(Block *)block forSelector:(SEL)aSelector withSignatureTypes:(NSString *)objCTypes
{
    [self setBlock:block forSelector:aSelector withSignature:[NSMethodSignature signatureWithObjCTypes:[objCTypes UTF8String]]];
}

-(void)setBlock:(Block *)block forSelector:(SEL)aSelector withSignature:(NSMethodSignature *)signature
{    
    // validate arguments
    // error modes:
    //    attempt to set block for native decorator method
    //    wrong number of arguments for block (wrt. selector)
    //    wrong number of arguments for signature (wrt. selector)
    if (block != nil)
    {
        if ([super respondsToSelector:aSelector])
        {
            FSExecError([NSString stringWithFormat:@"Cannot set a block for a method implemented by the decorator."]);
        }
        
        int blockArgCount = [block argumentCount];
        int selectorArgCount = [self argumentCountForSelector:aSelector];
        int signatureArgCount = [signature numberOfArguments] - 2; // don't want to count self and _cmd here
        
        if (blockArgCount - selectorArgCount != 1)
        {
            FSExecError([NSString stringWithFormat:@"Cannot set block %@ for #%@.  Wrong number of arguments. \
    (Possible cause: block to be bound to selector must take 'self' as first argument.)", block, NSStringFromSelector(aSelector)]);
        }
        
        if (signature != nil && selectorArgCount != signatureArgCount)
        {
            FSExecError([NSString stringWithFormat:@"Argument count of selector #%@ and signature %@ must match.", NSStringFromSelector(aSelector), signature]);       
        }
    }
    
    // go ahead and set
    NSValue *selKey = [NSValue value:&aSelector withObjCType:@encode(SEL)];
    if (block != nil)
    {
        [selectorsToBlocks setObject:block forKey:selKey];
        if (signature != nil)
            [selectorsToSignatures setObject:signature forKey:selKey];
        else
            [selectorsToSignatures removeObjectForKey:selKey];        
    }
    else
    {
        [selectorsToBlocks removeObjectForKey:selKey];
        [selectorsToSignatures removeObjectForKey:selKey];
    }
    
}


@end
